<?php

namespace Drupal\charts_kendo_ui\Plugin\chart\Library;

use Drupal\charts\Attribute\Chart;
use Drupal\charts\Element\Chart as ChartElement;
use Drupal\charts\Plugin\chart\Library\ChartBase;
use Drupal\charts\TypeManager;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormBuilder;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The 'Kendo UI.js' chart library attribute.
 */
#[Chart(
  id: "kendo_ui",
  name: new TranslatableMarkup("Kendo UI"),
  types: [
    "area",
    "arearange",
    "bar",
    "boxplot",
    "bubble",
    "candlestick",
    "column",
    "donut",
    "gauge",
    "heatmap",
    "line",
    "pie",
    "radar",
    "scatter",
    "spline",
  ]
)]
class KendoUi extends ChartBase implements ContainerFactoryPluginInterface {

  /**
   * The element info manager.
   *
   * @var \Drupal\Core\Render\ElementInfoManagerInterface
   */
  protected $elementInfo;

  /**
   * The chart type manager.
   *
   * @var \Drupal\charts\TypeManager
   */
  protected $chartTypeManager;

  /**
   * The chart type manager.
   *
   * @var \Drupal\Core\Form\FormBuilder
   */
  protected $formBuilder;

  /**
   * Constructs a \Drupal\views\Plugin\Block\ViewsBlockBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Render\ElementInfoManagerInterface $element_info
   *   The element info manager.
   * @param \Drupal\charts\TypeManager $chart_type_manager
   *   The chart type manager.
   * @param \Drupal\Core\Form\FormBuilder $form_builder
   *   The form builder.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ElementInfoManagerInterface $element_info, TypeManager $chart_type_manager, FormBuilder $form_builder) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->elementInfo = $element_info;
    $this->chartTypeManager = $chart_type_manager;
    $this->formBuilder = $form_builder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('element_info'),
      $container->get('plugin.manager.charts_type'),
      $container->get('form_builder'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function preRender(array $element) {
    $chart_definition = [];

    if (!isset($element['#id'])) {
      $element['#id'] = Html::getUniqueId('kendo_ui-render');
    }
    $chart_definition = $this->populateOptions($element, $chart_definition);
    $chart_definition = $this->populateAxes($element, $chart_definition);
    $chart_definition = $this->populateData($element, $chart_definition);

    $element['#attached']['library'][] = 'charts_kendo_ui/kendo_ui';
    $element['#attached']['library'][] = 'charts_kendo_ui/kendo_license';
    $element['#attributes']['class'][] = 'charts-kendo-ui';
    $element['#chart_definition'] = $chart_definition;

    return $element;
  }

  /**
   * Populate options.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  protected function populateOptions(array $element, array $chart_definition) {

    $chart_type = $this->chartTypeConversion($element['#chart_type']);

    if ($chart_type === 'line' && !empty($element['#polar'])) {
      $chart_type = 'radarLine';
    }
    if ($element['#chart_type'] === 'spline') {
      $chart_definition['stroke']['curve'] = 'smooth';
    }
    $chart_definition['seriesDefaults']['stack'] = (bool) $element['#stacking'];
    if (empty($element['#connect_nulls'])) {
      $chart_definition['seriesDefaults']['missingValues'] = 'gap';
    }
    else {
      $chart_definition['seriesDefaults']['missingValues'] = 'interpolate';
    }
    $chart_definition['seriesDefaults']['type'] = $chart_type;
    $chart_definition['title']['text'] = $element['#title'] ?? '';
    $chart_definition['subtitle']['text'] = $element['#subtitle'] ?? '';

    // Title position.
    if (!empty($element['#title_position'])) {
      $position_map = [
        'top' => 'top',
        'left' => 'top',
        'right' => 'bottom',
        'bottom' => 'bottom',
      ];
      $chart_definition['title']['position'] = $position_map[$element['#title_position']] ?? 'top';
    }

    // Tooltips configuration.
    if (isset($element['#tooltips'])) {
      $chart_definition['tooltip'] = [
        'visible' => (bool) $element['#tooltips'],
        'format' => '{0}',
        'shared' => TRUE,
      ];

      if (isset($element['#tooltips_use_html']) && $element['#tooltips_use_html']) {
        $chart_definition['tooltip']['template'] = '#= category #: #= value #';
      }
    }

    // Data labels configuration.
    if (isset($element['#data_labels']) && $element['#data_labels']) {
      $chart_definition['seriesDefaults']['labels'] = [
        'visible' => TRUE,
        'format' => '{0}',
        'background' => 'transparent',
      ];
    }

    // Data markers configuration.
    if (!empty($element['#data_markers'])) {
      $chart_definition['seriesDefaults']['markers'] = [
        'visible' => (bool) $element['#data_markers'],
        'size' => 8,
      ];
    }

    // Legend configuration.
    if (isset($element['#legend'])) {
      $chart_definition['legend'] = [
        'visible' => (bool) $element['#legend'],
      ];

      // Legend position.
      if (!empty($element['#legend_position'])) {
        $legend_position_map = [
          'top' => 'top',
          'right' => 'right',
          'bottom' => 'bottom',
          'left' => 'left',
        ];
        $chart_definition['legend']['position'] = $legend_position_map[$element['#legend_position']] ?? 'bottom';
      }

      // Legend title.
      if (!empty($element['#legend_title'])) {
        $chart_definition['legend']['title'] = [
          'text' => $element['#legend_title'],
          'visible' => TRUE,
        ];

        // Legend title styling.
        if (!empty($element['#legend_title_font_size'])) {
          $chart_definition['legend']['title']['font'] = "normal " .
            ($element['#legend_title_font_weight'] ?? 'normal') . " " .
            $element['#legend_title_font_size'] . "px " . ($element['#font'] ?? 'Arial');
        }
        if (!empty($element['#legend_title_font_style']) && $element['#legend_title_font_style'] === 'italic') {
          $chart_definition['legend']['title']['font'] = "italic " .
            ($chart_definition['legend']['title']['font'] ?? '');
        }
      }

      // Legend item styling.
      if (!empty($element['#legend_font_size'])) {
        $chart_definition['legend']['labels'] = [
          'font' => "normal " . ($element['#legend_font_weight'] ?? 'normal') . " " .
            $element['#legend_font_size'] . "px " . ($element['#font'] ?? 'Arial'),
        ];

        if (!empty($element['#legend_font_style']) && $element['#legend_font_style'] === 'italic') {
          $chart_definition['legend']['labels']['font'] = "italic " .
            ($chart_definition['legend']['labels']['font'] ?? '');
        }
      }
    }

    if ($element['#chart_type'] === 'gauge') {
      $chart_definition['fill']['gradient']['opacityFrom'] = (int) $element['#gauge']['min'];
      $chart_definition['fill']['gradient']['opacityTo'] = (int) $element['#gauge']['max'];
      $chart_definition['fill']['gradient']['type'] = 'horizontal';
      $stops = [];
      // If the following have a value, place them in the stops array.
      if ((int) $element['#gauge']['min'] === 0 || !empty($element['#gauge']['min'])) {
        $stops[] = (int) $element['#gauge']['min'];
      }
      if (!empty($element['#gauge']['yellow_from'])) {
        $stops[] = (int) $element['#gauge']['yellow_from'];
      }
      if (!empty($element['#gauge']['green_from'])) {
        $stops[] = (int) $element['#gauge']['green_from'];
      }
      if (!empty($element['#gauge']['max'])) {
        $stops[] = (int) $element['#gauge']['max'];
      }
      $chart_definition['fill']['gradient']['stops'] = $stops;
    }

    // Size and background.
    if (!empty($element['#width'])) {
      $chart_definition['chartArea']['width'] = $element['#width'] . ($element['#width_units'] ?? 'px');
    }
    if (!empty($element['#height'])) {
      $chart_definition['chartArea']['height'] = $element['#height'] . ($element['#height_units'] ?? 'px');
    }

    // Background color.
    if (!empty($element['#background'])) {
      $chart_definition['chartArea']['background'] = $element['#background'];
    }

    // Merge in chart raw options.
    if (!empty($element['#raw_options'])) {
      $chart_definition = NestedArray::mergeDeepArray([
        $chart_definition,
        $element['#raw_options'],
      ]);
    }

    return $chart_definition;
  }

  /**
   * Utility to populate data.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  protected function populateData(array &$element, array $chart_definition) {
    $categories = [];
    $chart_type = $this->chartTypeConversion($element['#chart_type']);
    // Loop through the Chart render elements looking for the xaxis, which
    // contains the labels.
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_xaxis' && !empty($element[$key]['#labels'])) {
        if (in_array($chart_type, ['pie', 'donut'])) {
          $categories = $element[$key]['#labels'];
          break;
        }
        $categories[] = $element[$key]['#labels'];
      }
    }
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_data') {
        $series = [];
        $series_data = [];

        // Make sure defaults are loaded.
        if (empty($element[$key]['#defaults_loaded'])) {
          $element[$key] += $this->elementInfo->getInfo($element[$key]['#type']);
        }

        // Convert target named axis keys to integers.
        if (isset($element[$key]['#target_axis'])) {
          $axis_name = $element[$key]['#target_axis'];
          foreach (Element::children($element) as $axis_key) {
            if ($element[$axis_key]['#type'] === 'chart_yaxis') {
              if ($axis_key === $axis_name) {
                break;
              }
            }
          }
          $series['axis'] = $axis_name;
        }
        elseif ($chart_definition['seriesDefaults']['type'] !== 'radarLine') {
          // If the chart type is a gauge, we need to set the axis to the main
          // axis.
          $series['axis'] = 'main';
        }

        // Allow data to provide the labels.
        // This will override the axis settings.
        if ($element[$key]['#labels'] && !in_array($element[$key]['#chart_type'], [
          'scatter',
          'bubble',
        ])) {
          foreach ($element[$key]['#labels'] as $label_index => $label) {
            $series_data[$label_index][0] = $label;
          }
        }
        elseif (!empty($categories) && in_array($chart_type, ['pie', 'donut'])) {
          foreach ($categories as $label_index => $label) {
            $series_data[$label_index]['category'] = $label;
            //$series_data[$label_index]['value'] = $element[$key]['#data'][$label_index];
          }
        }

        // Populate the data.
        foreach ($element[$key]['#data'] as $data_index => $data) {
          // Check if the format of the data is an array.
          $data_is_array = is_array($data);
          if ($data_is_array && !empty($categories) && in_array($chart_type, ['rangeArea', 'boxPlot', 'candlestick', 'heatmap'])) {
            $series_data[$data_index] = $data;
          }
          elseif (in_array($chart_type, ['donut', 'pie'])) {
            $series_data[$data_index]['value'] = $element[$key]['#data'][$data_index];
          }
          else {
            // Generally pie or donut charts, but not when grouping colors are
            // used.
            if (isset($series_data[$data_index])) {
              $series_data[$data_index][] = $data;
            }
            elseif (in_array($chart_type, ['bubble', 'gauge', 'scatter'])) {
              // @todo See about applying to pie and donut.
              $series_data[$data_index] = $data;
              $name = $series_data[$data_index]['name'] ?? NULL;
              if (!empty($name)) {
                $chart_definition['labels'][] = $name;
              }
              if (!empty($element[$key]['#grouping_colors'][$data_index][$name])) {
                $series_data[$data_index]['color'] = $element[$key]['#grouping_colors'][$data_index][$name];
              }
              elseif (!empty($element[$key]['#grouping_colors'][$data_index]) && is_array($element[$key]['#grouping_colors'][$data_index])) {
                $chart_definition['colors'][$data_index] = reset($element[$key]['#grouping_colors'][$data_index]);
              }
            }
            else {
              $series_data[$data_index] = [$data];
            }
          }
        }
        // In the case of combo charts, we need to pass the chart type at the
        // series level.
        $series['type'] = !empty($element[$key]['#chart_type']) ? $element[$key]['#chart_type'] : $element['#chart_type'];
        if ($series['type'] === 'spline') {
          $series['type'] = 'line';
          $series['style'] = 'smooth';
        }
        $series['type'] = $this->chartTypeConversion($series['type']);
        if ($element['#chart_type'] === 'donut') {
          // Add innerSize to differentiate between donut and pie.
          $series['innerSize'] = '40%';
        }
        $series['name'] = $element[$key]['#title'];
        $series['color'] = $element[$key]['#color'];
        if (in_array($series['type'], ['boxPlot', 'candlestick', 'line'])) {
          $chart_definition['stroke']['width'][] = 2;
        }
        else {
          $chart_definition['stroke']['width'][] = 0;
        }
        if ($series['type'] === 'line' && !empty($element['#polar'])) {
          unset($series['type']);
        }

        // Remove unnecessary keys to trim down the resulting JS settings.
        ChartElement::trimArray($series);
        // Assign the data to the series.
        $series['data'] = $series_data;

        // Merge in series raw options.
        if (!empty($element[$key]['#raw_options'])) {
          $series = NestedArray::mergeDeepArray([
            $series,
            $element[$key]['#raw_options'],
          ]);
        }
        // Handle the 'y_only' chart types defined in charts.charts_types.yml.
        if ($series['type'] === 'gauge') {
          // Check if $series['data'] is in the format [1, 2, 3].
          if (is_array($series['data'][0])) {
            $series_data = array_column($series['data'], 1);
          }
          else {
            // Remove all the unnecessary keys.
            $series_data = $series['data'];
          }
          // If the type is gauge, we need to set the fill colors.
          if ($element[$key]['#color']) {
            $chart_definition['fill']['colors'] = $element[$key]['#color'];
          }
          // Kendo UI expects a percentage for the gauge. If $series_data[0]
          // is greater than 100, we need to divide it by the max value.
          // The Charts style plugin is not configured to deliver more than
          // one series (it might be good to allow this in the future);
          // regardless, we need to accommodate charts built from the field
          // or the Charts API.
          foreach ($series_data as $series_key => $series_value) {
            if ($series_value > 100) {
              if (!empty($element['#gauge']['max'])) {
                // @todo Allow users to configure the precision.
                $percent = round($series_value / (int) $element['#gauge']['max'], 2) * 100;
                $series_data[$series_key] = $percent;
              }
            }
          }
          // Pass the series data into the chart definition.
          $chart_definition['series'] = $series_data;
          // If the series data is a one-dimensional array, we need to set the
          // labels to the series name. A good example would be for a gauge.
          if (empty(array_column($series['data'], 0))) {
            if (!empty($series['name'])) {
              $chart_definition['labels'][] = $series['name'];
            }
          }
          else {
            // If the series data is a two-dimensional array, we set the labels
            // with the first column of the series data.
            $chart_definition['labels'] = array_column($series['data'], 0);
          }
        }
        else {
          $chart_definition['series'][] = $series;
        }

      }
    }

    return $chart_definition;
  }

  /**
   * Populate axes.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  protected function populateAxes(array $element, array $chart_definition) {
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_xaxis' || $element[$key]['#type'] === 'chart_yaxis') {
        // Make sure defaults are loaded.
        if (empty($element[$key]['#defaults_loaded'])) {
          $element[$key] += $this->elementInfo->getInfo($element[$key]['#type']);
        }

        // Populate the chart data.
        $axis_type = $element[$key]['#type'] === 'chart_xaxis' ? 'xaxis' : 'yaxis';
        $axis = [];
        $axis['title']['text'] = $element[$key]['#title'];
        if (!empty($element[$key]['#labels']) && !in_array($element['#chart_type'], ['pie', 'donut'])) {
          $axis['categories'] = $element[$key]['#labels'];
        }
        if ($element[$key]['#opposite']) {
          $axis['name'] = $key;
          $chart_definition['categoryAxis']['axisCrossingValues'] = [0, 10];
        }
        if (!$element[$key]['#opposite']
          && $axis_type === 'yaxis'
          && $chart_definition['seriesDefaults']['type'] !== 'radarLine'
        ) {
          $axis['name'] = 'main';
        }

        // Merge in axis raw options.
        if (!empty($element[$key]['#raw_options'])) {
          $axis = NestedArray::mergeDeepArray([
            $axis,
            $element[$key]['#raw_options'],
          ]);
        }

        if ($axis_type === 'yaxis') {
          $chart_definition['valueAxis'][] = $axis;
        }
        else {
          $chart_definition['categoryAxis'] = $axis;
        }
      }
    }

    return $chart_definition;
  }

  /**
   * Convert module chart type conventions to ApexCharts conventions.
   *
   * @param string $chart_type
   *   The chart type.
   *
   * @return string
   *   The converted chart type.
   */
  protected function chartTypeConversion(string $chart_type): string {
    $chart_type_map = [
      'arearange' => 'rangeArea',
      'boxplot' => 'boxPlot',
      'spline' => 'line',
    ];

    return !empty($chart_type_map[$chart_type]) ? $chart_type_map[$chart_type] : $chart_type;
  }

}
